@@ -45,9 +45,10 @@ _SYSTEM_PREFIX = 'system.'
 class _RestClient:
     """A generic HTTP Rest client."""
 
-    def __init__(self, url: str) -> None:
+    def __init__(self, url: str, **connect_args) -> None:
         self._url = url
         self._con = None
+        self._connect_args = connect_args
 
     def __str__(self):
         return f'_RestClient for {self._url}'
@@ -61,6 +62,7 @@ class _RestClient:
 
     async def __aexit__(self, exc_type, exc, tb):
         await self._con.close()
+        self._connect_args = None
 
     async def connect(self):
         # set trust_env to True to enable env variable settings for network proxy support
@@ -68,6 +70,7 @@ class _RestClient:
 
     async def close(self):
         await self._con.close()
+        self._connect_args = None
 
     async def request(
         self,
@@ -90,20 +93,20 @@ class _RestClient:
 
         try:
             if method == 'GET':
-                return await self._con.get(url, params=query_param, headers=headers)
+                return await self._con.get(url, params=query_param, headers=headers, **self._connect_args)
             elif method == 'POST':
-                return await self._con.post(url, params=query_param, headers=headers, data=data)
+                return await self._con.post(url, params=query_param, headers=headers, data=data, **self._connect_args)
             elif method == 'DELETE':
-                return await self._con.delete(url, params=query_param, headers=headers)
+                return await self._con.delete(url, params=query_param, headers=headers, **self._connect_args)
             elif method == 'PUT':
-                return await self._con.put(url, params=query_param, headers=headers, data=data)
+                return await self._con.put(url, params=query_param, headers=headers, data=data, **self._connect_args)
         except Exception as e:
             raise VantiqException('io.vantiq.python.exception',
                                   'Unexpected exception performing {0} against server {1},',
                                   [method, self._url]) from e
 
     async def download(self, url: str, headers: Union[dict, None]) -> aiohttp.ClientResponse:
-        return await self._con.get(url, headers=headers)
+        return await self._con.get(url, headers=headers, **self._connect_args)
 
     async def upload(self, url: str, headers: Union[dict, None], content_type: str, filename: Union[str, None] = None,
                      doc_name: Union[str, None] = None,
@@ -118,7 +121,7 @@ class _RestClient:
             data.add_field(name=filename, value=inmem, content_type=content_type, filename=doc_name)
         else:
             data.add_field(name=filename, value=open(filename, 'rb'), content_type=content_type, filename=doc_name)
-        return await self._con.post(url, headers=headers, data=data)
+        return await self._con.post(url, headers=headers, data=data, **self._connect_args)
 
 
 class VantiqResources:
@@ -365,7 +368,7 @@ class Vantiq:
 
     """
 
-    def __init__(self, server: str, api_version: Union[str, None] = None):
+    def __init__(self, server: str, api_version: Union[str, None] = None, **connect_args):
         """Create a Vantiq client object.
 
         Parameters:
@@ -375,6 +378,9 @@ class Vantiq:
             api_version : str (optional)
                 Version of the API to use. Defaults to '1'
 
+            connect_args : the set of named connection arguments (optional)
+                These are passed through to the aiohttp and websockets connect calls
+
         Returns:
             Vantiq client object with which to interact with the Vantiq system.
 
@@ -393,6 +399,9 @@ class Vantiq:
                 ...
                 await client.close()
 
+            to pass special connection arguments
+            ::
+                client = Vantiq('https://dev.vantiq.com', '1', ssl=False)
 
         """
         self._vlog: Logger = logging.getLogger(self.__class__.__name__)
@@ -401,15 +410,18 @@ class Vantiq:
             self._server = server[:-len('/')]  # server.removesuffix('/')'
         else:
             self._server = server
+        self._connect_args = connect_args
         self._api_version = api_version if api_version else '1'
         self._is_authenticated = False
         self._username = None
+        self._impersonate = False
+        self._target_namespace = None
         self._access_token = None
         self._id_token = None
         self._auth_header = None
         self._connection = None
         self._is_connected = False
-        self._connection = _RestClient(self._server)
+        self._connection = _RestClient(self._server, **connect_args)
         self._base_path = '/api/v' + self._api_version + '/'
         self._subscriber: Union[_VantiqSubscriber, None] = None
 
@@ -427,6 +439,7 @@ class Vantiq:
 
     async def __aexit__(self, exc_type, exc, tb):
         await self.close()
+        self._connect_args = {}
 
     async def connect(self) -> None:
         """Make the connection to the Vantiq server
@@ -473,9 +486,18 @@ class Vantiq:
         """Returns the access token currently in use."""
         return self._access_token
 
+    def set_target_namespace(self, namespace: Union[str, None]) -> None:
+        """Set the target namespace to be used to access the Vantiq server"""
+        self._target_namespace = namespace
+
+    def get_target_namespace(self) -> Union[str, None]:
+        """Returns the target namespace currently in use."""
+        return self._target_namespace
+
     def set_username(self, username: str) -> None:
         """Set the username to be used to access the Vantiq server"""
         self._username = username
+        self._impersonate = True
 
     def get_username(self) -> str:
         """Returns the username currently in use."""
@@ -553,7 +575,9 @@ class Vantiq:
         self._id_token = None
         self._base_path = None
         self._auth_header = None
+        self._target_namespace = None
         self._username = None
+        self._connect_args = {}
         if self._subscriber:
             await self._subscriber.close()
             self._subscriber = None
@@ -594,12 +618,21 @@ class Vantiq:
                 params = []
             raise VantiqException(code, message, params)
 
+    def _get_auth_headers(self) -> dict:
+        headers = {aiohttp.hdrs.AUTHORIZATION: self._auth_header}
+        if self._target_namespace is not None:
+            headers['X-Target-Namespace'] = self._target_namespace
+        if self._username is not None and self._impersonate:
+            headers['X-Impersonate-User'] = self._username
+        return headers
+
     async def _perform_operation(self, operation: str, method: str, path: str,
                                  query_params: Union[dict, None], is_streaming: bool,
-                                 instance: Union[dict, None] = None) -> VantiqResponse:
+                                 instance: Union[dict, None] = None, headers: Union[dict, None] = None) -> VantiqResponse:
         if self._is_authenticated:
             try:
-                headers = {aiohttp.hdrs.AUTHORIZATION: self._auth_header}
+                headers = headers or {}
+                headers.update(self._get_auth_headers())
                 if instance is None:
                     # When no parameters are passed at all, we get 404's back.  So None as parameters == {}.
                     instance = {}
@@ -924,7 +957,7 @@ class Vantiq:
                         ...
         """
         url = path
-        headers = {aiohttp.hdrs.AUTHORIZATION: self._auth_header}
+        headers = self._get_auth_headers()
         try:
             resp = await self._connection.download(url, headers)
             ret_val = VantiqResponse(resp.ok, resp.status, resp.content_type)
@@ -1001,7 +1034,7 @@ class Vantiq:
             return VantiqResponse.from_error(ve)
 
         path = self._build_path(resource, None)
-        headers = {aiohttp.hdrs.AUTHORIZATION: self._auth_header}
+        headers = self._get_auth_headers()
 
         try:
             resp = await self._connection.upload(path, headers, content_type, filename, doc_name, inmem)
@@ -1084,7 +1117,7 @@ class Vantiq:
                                   'Unexpected error during {0} operation.',
                                   [operation]) from e
 
-    async def execute(self, procedure_id: str, params: dict) -> VantiqResponse:
+    async def execute(self, procedure_id: str, params: dict, headers: Union[dict, None] = None) -> VantiqResponse:
         """(Async) Execute a Vantiq procedure.
 
         Parameters:
@@ -1103,7 +1136,7 @@ class Vantiq:
             query_params = {}
             method = 'POST'
             path = self._build_path(_SYSTEM_PREFIX + 'procedures', procedure_id)
-            resp = await self._perform_operation(operation, method, path, query_params, False, params)
+            resp = await self._perform_operation(operation, method, path, query_params, False, params, headers)
             return resp
         except VantiqException:
             # If we've already handled or wrapped it, just pass it along
@@ -1222,7 +1255,8 @@ class Vantiq:
             return None
 
     async def subscribe(self, resource: str, resource_id: str, operation: Union[str, None],
-                        callback: Callable[[str, dict], Awaitable[None]], params: dict = None) -> VantiqResponse:
+                        callback: Callable[[str, dict], Awaitable[None]], params: Union[dict, None] = None,
+                        target_namespace: Union[str, None] = None) -> VantiqResponse:
         """(Async) Subscribe to an event from the Vantiq server.
 
         Subscribes to a specific topic, source, service, or type event.
@@ -1259,6 +1293,8 @@ class Vantiq:
 
             params : dict
                 (optional) Parameters for the subscription. May be subscription dependent, but can usually be ignored.
+            target_namespace : str
+                (optional) The namespace in which to subscribe.  If not provided, the current namespace is used.
         Returns:
             VantiqResponse indicating the success of the operation.
 
@@ -1289,7 +1325,7 @@ class Vantiq:
         if self._subscriber is None:
             await self.start_subscriber_transport()
 
-        vr = await self._subscriber.subscribe(path, params, callback)
+        vr = await self._subscriber.subscribe(path, params, callback, target_namespace)
         return vr
 
     async def ack(self, request_id: str, subscription_id: str, msg: dict) -> None:
@@ -1320,22 +1356,35 @@ class Vantiq:
         await self._subscriber.ack(request_id, subscription_id, sequence_id, partition_id)
         return
 
+    def register_subscriber_on_close(self, callback: Callable[[], Awaitable[None]]):
+        """ Register a callback to be called when the subscriber is closed.
+
+        Parameters:
+            callback : Callable[[], Awaitable[None]]
+                The callback to be called when the subscriber is closed.
+        """
+        self._subscriber.on_close_handler = callback
+
 
 class _VantiqSubscriber:
     CONNECT = 'connect'
     MESSAGE = 'message'
     ERROR = 'error'
 
-    def __init__(self, parent: Vantiq):
-        self.parent = parent
+    def __init__(self, parent: Vantiq, **connect_args):
+        self.parent: Vantiq = parent
+        self._connect_args = connect_args
         self.connected = False
         self.connected_future: asyncio.Future = asyncio.get_running_loop().create_future()
+        # noinspection PyTypeChecker
         self.connection: websockets.ClientProtocol = None
         self.url = None
         self._vlog = logging.getLogger(self.__class__.__name__)
         self.subscriptions: Dict[str, bool] = {}
         self.callbacks: Dict[str, Callable[[str, dict], Awaitable[None]]] = {}
         self.is_authenticated = False
+        # noinspection PyTypeChecker
+        self.on_close_handler: Callable[[], Awaitable[None]] = None
 
     def __str__(self):
         ret_val = f'VantiqSubscriber for {str(self.parent)}'
@@ -1360,94 +1409,81 @@ class _VantiqSubscriber:
                   "/wsock/websocket"
             async with websockets.connect(uri=self.url,
                                           ping_interval=20 if do_pings else None,
-                                          ping_timeout=20 if do_pings else None) as websocket:
-                if not self.connected:
-                    auth_msg = {
-                        'op': 'validate',
-                        'resourceName': 'system.credentials',
-                        'object': self.parent.get_access_token()
-                    }
-                    await websocket.send(json.dumps(auth_msg))
+                                          ping_timeout=20 if do_pings else None,
+                                          **self._connect_args) as websocket:
+                try:
+                    if not self.connected:
+                        auth_msg = {
+                            'op': 'validate',
+                            'resourceName': 'system.credentials',
+                            'object': self.parent.get_access_token()
+                        }
+                        await websocket.send(json.dumps(auth_msg))
 
-                async for raw in websocket:
+                    # Read from the socket until it closes
+                    async for raw in websocket:
+                        # Decode the message
+                        resp = json.loads(raw)
+                        request_id = None
+                        if 'headers' in resp.keys():
+                            hdrs = resp['headers']
+                            if 'X-Request-Id' in hdrs.keys():
+                                request_id = hdrs['X-Request-Id']
+                        else:
+                            pass
 
-                    resp = json.loads(raw)
-                    request_id = None
-                    if 'headers' in resp.keys():
-                        hdrs = resp['headers']
-                        if 'X-Request-Id' in hdrs.keys():
-                            request_id = hdrs['X-Request-Id']
-                    else:
-                        pass
-
-                    # Using request id, track back to request for which this is a response
-
-                    if 'status' in resp.keys():
-                        if resp['status'] == 200:
-                            if not self.is_authenticated:
-                                self.connection = websocket
-                                self.connected = True
-                                self.is_authenticated = True
-                                self.connected_future.set_result('OK')
-                                self._vlog.debug('Authentication completed.')
-                            else:
-                                if request_id:
-                                    if request_id in self.subscriptions.keys():
-                                        if not self.subscriptions[request_id]:
-                                            # Then this is our response for our subscription.
-                                            self.subscriptions[request_id] = True
-                                            self._vlog.debug('Subscription requested accepted.')
-                                            callback = self.callbacks[request_id]
-                                            await callback(self.CONNECT, resp)
-                                        elif request_id in self.callbacks.keys():
-                                            callback = self.callbacks[request_id]
-                                            await callback(self.MESSAGE, resp)
-                        elif resp['status'] >= 400:
-                            if not self.connected:
-                                self._vlog.error('Connect call failed: %s :: %s:%s', resp['status'],
-                                                 resp['body'][0]['code'], resp['body'][0]['message'])
-                                ve = VantiqException('io.vantiq.python.connect.failed',
-                                                     'Connect call failed: {0} :: {1}:{2}',
-                                                     [resp['status'], resp['body'][0]['code'],
-                                                      resp['body'][0]['message']])
-                                self.connected_future.set_exception(ve)
-                                raise ve
-                            else:
+                        # Using request id, track back to request for which this is a response
+                        if 'status' in resp.keys():
+                            if resp['status'] == 200:
+                                if not self.is_authenticated:
+                                    self.connection = websocket
+                                    self.connected = True
+                                    self.is_authenticated = True
+                                    self.connected_future.set_result('OK')
+                                    self._vlog.debug('Authentication completed.')
+                                else:
+                                    if request_id:
+                                        if request_id in self.subscriptions.keys():
+                                            if not self.subscriptions[request_id]:
+                                                # Then this is our response for our subscription.
+                                                self.subscriptions[request_id] = True
+                                                self._vlog.debug('Subscription requested accepted.')
+                                                callback = self.callbacks[request_id]
+                                                await callback(self.CONNECT, resp)
+                                            elif request_id in self.callbacks.keys():
+                                                callback = self.callbacks[request_id]
+                                                body = resp.get('body', None)
+                                                if body is not None and isinstance(body, dict) and body.get('op', None) == 'unsubscribe':
+                                                    self.subscriptions.pop(request_id)
+                                                    self.callbacks.pop(request_id)
+                                                await callback(self.MESSAGE, resp)
+                            elif resp['status'] >= 400:
+                                if not self.connected:
+                                    self._vlog.error('Connect call failed: %s :: %s:%s', resp['status'],
+                                                     resp['body'][0]['code'], resp['body'][0]['message'])
+                                    ve = VantiqException('io.vantiq.python.connect.failed',
+                                                         'Connect call failed: {0} :: {1}:{2}',
+                                                         [resp['status'], resp['body'][0]['code'],
+                                                          resp['body'][0]['message']])
+                                    self.connected_future.set_exception(ve)
+                                    raise ve
+                                else:
+                                    if request_id and request_id in self.callbacks.keys():
+                                        callback = self.callbacks[request_id]
+                                        await callback(self.ERROR, resp)
+                            elif resp['status'] == 100:
                                 if request_id and request_id in self.callbacks.keys():
+                                    self._vlog.debug('Message received via subscription.')
                                     callback = self.callbacks[request_id]
-                                    await callback(self.ERROR, resp)
-                        elif resp['status'] == 100:
-                            if request_id and request_id in self.callbacks.keys():
-                                self._vlog.debug('Message received via subscription.')
-                                callback = self.callbacks[request_id]
-                                await callback(self.MESSAGE, resp)
-            self.connected = False
+                                    await callback(self.MESSAGE, resp)
+                finally:
+                    # Once we get here, any transient subscriptions will be gone, so we should reset our side as well
+                    await self.unsubscribe_all()
 
-    async def subscribe(self, path: str, params: dict,
-                        callback: Callable[[str, dict], Awaitable[None]]) -> VantiqResponse:
-        if self.connected:
-            if path in self.subscriptions.keys():
-                vr = VantiqResponse(False, 400, None)
-                ve = VantiqError('io.vantiq.python.subscribed',
-                                 'A subscription for path {0} already exists.',
-                                 [path])
-                vr.errors = [ve]
-                return vr
-            self.callbacks[path] = callback
-            self.subscriptions[path] = False  # Will be set to true when the server responds
-            if params is None:
-                params = {}
-            params['requestId'] = path
-            sub_msg = {'op': 'subscribe',
-                       'resourceName': 'events',
-                       'resourceId': path,
-                       'parameters': params}
-            await self.connection.send(json.dumps(sub_msg))
-            self._vlog.debug('Subscription request sent.')
-
-            vr = VantiqResponse(True, 204, None)
-            return vr
-        else:
+    async def subscribe(self, path: str, params: dict, callback: Callable[[str, dict], Awaitable[None]],
+                        target_namespace: Union[str, None] = None) -> VantiqResponse:
+        # If we aren't connected, complain
+        if not self.connected:
             self._vlog.error('No transport for subscriptions established.')
             vr = VantiqResponse(False, 400, None)
             ve = VantiqError('io.vantiq.python.notransport',
@@ -1456,12 +1492,45 @@ class _VantiqSubscriber:
             vr.errors = [ve]
             return vr
 
+        # Construct the request id using the path and (optional) target namespace
+        target_namespace = target_namespace or self.parent.get_target_namespace()
+        request_id = path
+        if target_namespace is not None:
+            request_id += '@' + target_namespace
+
+        # Check to see if we already have a subscription for this path
+        if self.subscriptions.get(request_id, False):
+            vr = VantiqResponse(False, 400, None)
+            ve = VantiqError('io.vantiq.python.subscribed',
+                             'A subscription for path {0} already exists for ns {1}.',
+                             [path, target_namespace or "current"])
+            vr.errors = [ve]
+            return vr
+        self.callbacks[request_id] = callback
+        self.subscriptions[request_id] = False  # Will be set to true when the server responds
+        if params is None:
+            params = {}
+        params['requestId'] = request_id
+        sub_msg = {'op': 'subscribe',
+                   'resourceName': 'events',
+                   'resourceId': path,
+                   'parameters': params}
+        if target_namespace is not None:
+            sub_msg['targetNamespace'] = target_namespace
+        # noinspection PyUnresolvedReferences
+        await self.connection.send(json.dumps(sub_msg))
+        self._vlog.debug('Subscription request sent.')
+        return VantiqResponse(True, 204, None)
+
     async def ack(self, request_id: str, subscription_id: str, sequence_id: float, partition_id: float):
         params = {'requestId': request_id,
                   'subscriptionName': subscription_id, 'sequenceId': sequence_id, 'partitionId': partition_id}
         msg = {'op': 'acknowledge', 'resourceName': 'events', 'resourceId': request_id,
                'parameters': params}
+        if self.parent.get_target_namespace() is not None:
+            msg['targetNamespace'] = self.parent.get_target_namespace()
         raw = json.dumps(msg)
+        # noinspection PyUnresolvedReferences
         await self.connection.send(raw)
 
     async def unsubscribe_all(self):
@@ -1471,6 +1540,10 @@ class _VantiqSubscriber:
 
     async def close(self):
         self.connected = False
+        # noinspection PyUnresolvedReferences
         await self.connection.close()
         self.connection = None
+        self._connect_args = {}
         self.is_authenticated = False
+        if self.on_close_handler is not None:
+            await self.on_close_handler()
